import os
import yaml
import logging
import riemann_client.transport

from functools import partial

from riemann_client.transport import BlankTransport, TLSTransport, TCPTransport

from jinja2 import Template
from .util import dynamic_import


class ConfigBase(object):

    """
    Base config class
    """

    def is_valid(self):
        return True


class InstanceMixin(object):
    """
    Mixin used for configs which are loaded dynamically
    """
    @property
    def module(self):
        return self._data["module"]

    @property
    def instance(self):
        if self._instance is None:
            self._instance = dynamic_import(self.module)(self._data)
        return self._instance

    def is_valid(self):
        return self.instance.is_valid()


class TagMixin(object):
    """
    Mixin used for adding tag functionallity to configs
    """

    @property
    def tag(self):
        return self._data.get("tag", None)

    @property
    def tags(self):
        return self._data.get("tags", []) + ([self.tag] if self.tag else [])


class RiemannConfig(ConfigBase):

    """
    Riemann Config object
    """

    def __init__(self, cfg):
        self._data = cfg

    @property
    def host(self):
        return self._data.get("host", "localhost")

    @property
    def port(self):
        return int(self._data.get("port", 5555))

    @staticmethod
    def default():
        return RiemannConfig({})

    @property
    def ca_certs(self):
        return self._data.get("ca-certs", None)

    def get_transport(self, noop=False):
        if noop:
            return BlankTransport
        else:
            return self.transport

    @property
    def transport(self):
        raw = self._data.get("transport", None)
        if raw:  # Transport is defined
            return getattr(riemann_client.transport, raw)
        elif self.ca_certs:
            return partial(TLSTransport, ca_certs=self.ca_certs)
        else:
            return TCPTransport


class AgentConfig(InstanceMixin, TagMixin):

    """
    Config for setuping agent
    """

    def __init__(self, cfg):
        self._data = cfg
        self._instance = None


class AugmentConfig(InstanceMixin, TagMixin):

    """
    Config for setuping augment
    """

    def __init__(self, cfg):
        self._data = cfg
        self._instance = None

    @property
    def key(self):
        return self._data["key"]


class Config(ConfigBase):

    """
    Config object
    """

    def __init__(self, cfg):
        self._data = cfg
        self.is_valid()

    def is_valid(self):
        checks = ([self.riemann.is_valid] +
                  list(map(lambda x: x.is_valid, self.agents)))
        return all([c() for c in checks])

    @property
    def riemann(self):
        if "riemann" in self._data:
            return RiemannConfig(self._data["riemann"])
        else:
            return RiemannConfig.default()

    @property
    def interval(self):
        return self._data["interval"]

    @property
    def agents(self):
        return [AgentConfig(a) for a in self._data.get("agents", [])]

    @property
    def augments(self):
        return [AugmentConfig(a) for a in self._data.get("augments", [])]

    @property
    def sentry_dsn(self):
        return self._data.get("sentry-dsn", None)

    @property
    def executors_count(self):
        return self._data.get("executors-count", 10)

    @property
    def executor_class(self):
        raw = self._data.get(
                "executor",
                "concurrent.futures.ThreadPoolExecutor"
        )
        return dynamic_import(raw)


def load(config_file):
    """
    Processes and loads config file.
    """
    with open(config_file, "r") as f:

        def env_get():
            return dict(os.environ)
        tmpl = Template(f.read())
        return Config(yaml.safe_load(tmpl.render(**env_get())))



from time import time

from oshino.config import TagMixin


class AugmentBase(TagMixin):
    def __init__(self, data):
        self._data = data

    def activate(self, client, g):
        pass

    def is_valid(self):
        return True

    @property
    def name(self):
        return self._data["name"].lower().replace(" ", "-")

    @property
    def prefix(self):
        return "{0}.".format(self.name)

    @property
    def key(self):
        return self._data["key"]

    def send_event(self, client, **kwargs):
        tags = self.tags
        if "tags" in kwargs:
            for tag in tags:
                kwargs["tags"].append(tag)
        else:
            kwargs["tags"] = tags

        if "time" not in kwargs:
            kwargs["time"] = int(time())

        client.event(**kwargs)


class InvalidAugment(AugmentBase):
    """
    Only for Testing purposes
    """

    def is_valid(self):
        return False

    def activate(self, client, g):
        raise RuntimeError("We shouldn't get here!")



import re
import json
import asyncio

from functools import partial
from . import Agent


class SubprocessAgent(Agent):

    @property
    def script(self):
        return self._data["script"]

    def is_valid(self):
        return (super(SubprocessAgent, self).is_valid()
                and "script" in self._data)

    async def process(self, event_fn):
        proc = await asyncio.create_subprocess_shell(self.script)
        exitcode = await proc.wait()
        state = "ok" if exitcode == 0 else "failure"
        event_fn(service=self.prefix + "shell",
                 state=state,
                 metric_f=1.0,
                 description="Exit code: {0}".format(exitcode)
                 )


def split_transform(m, logger, metric_sep="="):

    try:
        key, val = m.split(metric_sep, 1)
        return key.strip(), float(val.strip())
    except Exception as ex:
        logger.warn("Failed to parse '{0}' with '{1}' as sep, Reason: {2}"
                    .format(m, metric_sep, ex))
        return None


def regex_transform(m, logger):
    raw = re.match(
            r"(?P<key>(\w|[.-])+)\s*[:=]\s*(?P<val>(-?\d+([.]\d{1,2})?))",
            m
    )

    if raw:
        return raw.group("key"), float(raw.group("val"))
    else:
        return None


def json_transform(m, logger, selector):
    raw = json.loads(m)
    return selector, raw[selector]


def is_parsed(m):
    return m is not None


class StdoutAgent(Agent):
    @property
    def cmd(self):
        return self._data["cmd"]

    @property
    def args(self):
        return self._data.get("args", {})

    @property
    def transform_fn(self):
        local_fn = self._data.get("local-transform", None)
        if local_fn:
            raw = "oshino.agents.subprocess_agent.{0}".format(local_fn)
        else:
            raw = self._data.get(
                    "transform-fn",
                    "oshino.agents.subprocess_agent.split_transform"
            )

        mod, func = raw.rsplit(".", 1)
        m = __import__(mod, fromlist=[func])
        return getattr(m, func)

    async def process(self, event_fn):
        logger = self.get_logger()
        proc = await asyncio.create_subprocess_shell(
                self.cmd,
                stdout=asyncio.subprocess.PIPE
        )

        def strip(x):
            return x.strip()
        stdout, stderr = await proc.communicate()
        content = stdout.decode().strip()
        transform = partial(self.transform_fn, logger=logger, **self.args)
        metrics = filter(
                is_parsed,
                map(transform, map(strip, content.split("\n")))
        )

        for key, val in metrics:
            event_fn(service=self.prefix + key,
                     state="ok",
                     metric_f=float(val))



import sys
import time

from datetime import datetime


def dynamic_import(path):
    module, builder = path.rsplit(".", 1)
    return getattr(__import__(module, fromlist=[builder]), builder)


def current_ts():
    """
    Just gives current timestamp.
    """
    utcnow = datetime.utcnow()
    return int(utcnow.timestamp() * 1000)


def timer():
    """
    Timer used for calculate time elapsed
    """
    if sys.platform == "win32":
        default_timer = time.clock
    else:
        default_timer = time.time

    return default_timer()



import logging

logger = logging.getLogger(__name__)


def send_heartbeat(event_fn, ttl):
    logger.debug("Sending heartbeat")
    event_fn(metric_f=1.0,
             service="oshino.heartbeat",
             tags=["oshino", "heartbeat"],
             ttl=ttl)


def send_timedelta(event_fn, td, interval):
    logger.debug("Oshino took: {0}ms to collect metrics"
                 .format(int(td * 1000)))
    if td > interval:
        state = "error"
    else:
        state = "ok"

    event_fn(metric_f=td,
             state=state,
             service="oshino.processing_time",
             tags=["oshino", "instrumentation"])


def send_metrics_count(event_fn, count):
    logger.debug("Oshino collected {0} metrics".format(count))
    event_fn(metric_f=count,
             service="oshino.metrics_count",
             tags=["oshino", "instrumentation"])


def send_pending_events_count(event_fn, count):
    logger.debug("There are {0} events which are still being processed"
                 .format(count))
    event_fn(metric_f=count,
             service="oshino.pending_events_count",
             tags=["oshino", "instrumentation"])



from collections import deque

from . import AugmentBase


class SimpleMovingAverage(AugmentBase):

    def __init__(self, *args, **kwargs):
        super(SimpleMovingAverage, self).__init__(*args, **kwargs)
        self.dq = deque(maxlen=self.step)

    @property
    def step(self):
        return self._data.get("step", 5)

    def activate(self, client, g):
        step = self.step

        stopped = False
        while not stopped:

            try:
                event = next(g)
                self.dq.append(event)
            except StopIteration:
                stopped = True
                break

            data = list(map(lambda x: x["metric_f"], list(self.dq)))
            if len(data) == step:
                output = sum(data) / len(data)

                self.send_event(client,
                                service=self.prefix + "value",
                                metric_f=output,
                                state="ok")



import logging
import asyncio

from copy import copy
from time import sleep

from queue import Queue

from riemann_client.client import QueuedClient
from riemann_client.transport import RiemannError
from riemann_client import riemann_pb2


logger = logging.getLogger("Processor")


class StopEvent(object):
    pass


class AugmentFixture(object):

    def apply_augment(self, data):
        if "service" not in data or hasattr(self, "stopped"):
            return

        key = data["service"]
        if key not in self.augments:
            return

        subscribers = self.augments[key]
        for sub in subscribers:
            sub.put_nowait(data)

    def on_stop(self):
        if hasattr(self, "stopped"):
            return

        logger.info("Stopping all augments")

        self.stopped = True
        for _, subscribers in self.augments.items():
            for sub in subscribers:
                sub.put_nowait(StopEvent())

        for _, subscribers in self.augments.items():
            for sub in subscribers:
                while not sub.empty():
                    sleep(0.1)


class QClient(QueuedClient, AugmentFixture):
    def __init__(self, *args, **kwargs):
        self.augments = {}
        self.tasks = []
        self.queue = riemann_pb2.Msg()

    def event(self, **data):
        self.tasks.append(self.apply_augment(copy(data)))
        super(QClient, self).event(**data)

    def clear_queue(self):
        self.queue = riemann_pb2.Msg()

    def flush(self, transport):
        response = transport.send(self.queue)
        self.clear_queue()
        return response


async def flush(client, transport):
    future = asyncio.Future()

    async def process_async(future):
        try:
            logger.debug("Flushing queue")
            transport.connect()
            client.flush(transport)
        except (RiemannError, ConnectionRefusedError) as ex:
            logger.warning(ex)
            future.set_result(False)
            logger.debug("Flush failed due to connection errors")
        except Exception as ex:
            logger.exception(ex)
            future.set_result(False)
            raise ex
        else:
            logger.debug("Flush sucessful")
            future.set_result(True)
        finally:
            transport.disconnect()

    asyncio.ensure_future(process_async(future))
    await future
    return future.result()


def register_augment(client, key, augment_fn):
    if key not in client.augments:
        client.augments[key] = []

    loop = asyncio.get_event_loop()

    def generator(q):
        stopped = False
        while not (stopped and q.empty()):
            logger.debug("Waiting for event for {0}".format(augment_fn))
            event = q.get(block=not stopped)
            if isinstance(event, StopEvent):
                logger.debug("Stopping event generator")
                stopped = True
                continue
            yield event
        logger.info("Shutting down generator")
        raise StopIteration

    q = Queue()
    g = generator(q)

    def execute_in_thread(fn, client, g):
        fn(client, g)

    loop.run_in_executor(None, execute_in_thread, augment_fn, client, g)

    client.augments[key].append(q)



import logging

from abc import ABC, abstractmethod

from oshino.util import current_ts, timer


class Agent(ABC):

    def __init__(self, cfg):
        self._data = cfg
        self._last_run = current_ts()

    def on_start(self):
        pass

    def on_stop(self):
        pass

    @property
    def name(self):
        return self._data["name"].lower().replace(" ", "-")

    @property
    def prefix(self):
        return "{0}.".format(self.name)

    def get_logger(self):
        return logging.getLogger(self.__class__.__name__)

    @abstractmethod
    async def process(self, event_fn):
        """
        Each agent must implement this one to provide actual logic
        """
        pass

    def is_valid(self):
        return "name" in self._data

    async def pull_metrics(self, event_fn, loop=None):
        """
        Method called by core.
        Should not be overwritten.
        """
        if self.lazy and not self.ready:
            return None
        logger = self.get_logger()

        ts = timer()
        # logger.trace("Waiting for process event")
        result = await self.process(event_fn)
        td = int(timer() - ts)
        # logger.trace("It took: {}ms".format(td))
        self._last_run = current_ts()
        return result

    @property
    def lazy(self):
        """
        Agents with flag `lazy` gives data when they want to,
        not when they are requested for.
        """
        return self._data.get("lazy", False)

    @property
    def ready(self):
        """
        Function used when agent is `lazy`.
        It is being processed only when `ready` condition is satisfied
        """
        logger = self.get_logger()
        now = current_ts()
        # logger.trace("Current time: {0}".format(now))
        # logger.trace("Last Run: {0}".format(self._last_run))
        delta = (now - self._last_run)
        # logger.trace("Delta: {0}, Interval: {1}"
        #              .format(delta, self.interval * 1000))
        return delta > self.interval * 1000

    @property
    def interval(self):
        """
        By default, lazy agents expects to have some kind of time interval.
        `ready` output is calculated according to this interval.
        """
        return self._data["interval"]



import logging

from argparse import ArgumentParser

from dotenv import load_dotenv, find_dotenv
from .config import load
from .core.heart import start_loop

logger = logging.getLogger(__name__)


try:
    load_dotenv(find_dotenv())
except Exception as ex:
    logger.error("Error while loading .env: '{}'. Ignoring.".format(ex))


def main(args):
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    cfg = load(args.config)
    start_loop(cfg, args.noop)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--config", help="Config file", default="config.yaml")
    parser.add_argument("--noop", action="store_true", default=False, help="Events will be processed, but not sent to Riemann")
    parser.add_argument("--debug", action="store_true", default=False, help="Debug mode")
    main(parser.parse_args())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['oshino', 'oshino.agents', 'oshino.augments', 'oshino.core']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.0.3,<4.0.0',
 'PyYAML>=6.0,<7.0',
 'aiohttp>=3.8.1,<4.0.0',
 'oshino-admin>=0.2.0,<0.3.0',
 'protobuf==3.1.0.post1',
 'pytest-asyncio>=0.17.2,<0.18.0',
 'python-dotenv>=0.19.2,<0.20.0',
 'raven>=6.10.0,<7.0.0',
 'requests>=2.27.1,<3.0.0',
 'riemann-client==6.3.0']

setup_kwargs = {
    'name': 'oshino',
    'version': '0.4.0',
    'description': '',
    'long_description': "About\n=====\nOshino - named after character Meme Oshino from anime called Bakemonogatari:\n> Meme Oshino (忍野 メメ, Oshino Meme) is a middle-aged man who lives with the mysterious Shinobu Oshino in an abandoned cram school building in the town Koyomi Araragi resides in Bakemonogatari. An expert in the supernatural, he is the reason why Koyomi was able to return back to normal after being bitten by a vampire, and he becomes Koyomi's informant when it comes to oddities for some time.\n[Source](https://myanimelist.net/character/22552/Meme_Oshino)\n\nJust like anime character, this service likes to deal with supernatural - system availability.\n\nHeavily inspired by [collectd](https://github.com/collectd/collectd) and\n[riemann-tools](https://github.com/riemann/riemann-tools), and unintentionally similar to [python-diamond](https://github.com/python-diamond/Diamond)\n\nAlerting and Monitoring based on [Riemann](https://riemann.io)\n\n\n[![Build Status](https://travis-ci.org/CodersOfTheNight/oshino.svg?branch=master)](https://travis-ci.org/CodersOfTheNight/oshino)\n[![Coverage Status](https://coveralls.io/repos/github/CodersOfTheNight/oshino/badge.svg?branch=master)](https://coveralls.io/github/CodersOfTheNight/oshino?branch=master)\n[![Documentation Status](https://readthedocs.org/projects/oshino/badge/?version=latest)](http://oshino.readthedocs.io/projects/https://github.com/CodersOfTheNight/oshino-consul/en/latest/?badge=latest)\n\n\nRequirements\n============\n- Python 3.5+ version\n- Have Riemann node running\n\nHow to install\n==============\n`pip install oshino`\n\nQuickstart\n==========\nIt is highly recommended for new users to use [Quickstart Guide](docs/quickstart.md)\n\n\nRiemann. What? Why? How?\n=========================\nRiemann is a backbone of this system. It does alerting, it receives metrics, it aggregates metrics and it decides where to send them (eg.: Graphite, Logstash).\nHowever, it is rather unknown to the most of people, and configuring can be not trivial at all. \n\nTo mitigate this problem, documentation for setuping Riemann for this scenario has been made:\n[riemann](docs/riemann.md)\n\nExample config\n--------------\n```yaml\n---\ninterval: 10\nriemann:\n  host: localhost\n  port: 5555\nagents:\n  - name: health-check\n    module: oshino.agents.http_agent.HttpAgent\n    url: http://python.org\n    tag: healthcheck\n```\n\nCustom Agents\n===============\nDocumentation about additional agents can be found [here](docs/thirdparty.md)\n\nMore documentation\n==================\nMore documentation can be found under [docs](docs/index.md) directory\n\nContributing\n============\nRefer to [CONTRIBUTING.md](CONTRIBUTING.md)\n",
    'author': 'Šarūnas Navickas',
    'author_email': 'zaibacu@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/CodersOfTheNight/oshino',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import logging
import asyncio
import sys

from typing import TypeVar, Generic
from asyncio import BaseEventLoop

from raven.handlers.logging import SentryHandler
from raven import Client as SentryClient
from raven.exceptions import InvalidDsn

from ..config import Config
from ..version import get_version
from ..util import timer
from . import (send_heartbeat,
               send_timedelta,
               send_pending_events_count,
               send_metrics_count)
from . import processor

logger = logging.getLogger(__name__)


T = TypeVar("T")


def create_loop():
    return asyncio.get_event_loop()


def forever():
    return True


def create_agents(agents_cfg: list):
    return list(map(lambda x: (x.instance, x),
                    filter(lambda x: x.is_valid(), agents_cfg)))


def register_augments(client: processor.QClient,
                      augments_cfg: list):
    for augment in augments_cfg:
        if not augment.is_valid():
            logger.warn("Augment '{0}' failed to pass validation"
                        .format(augment))
            continue

        inst = augment.instance

        processor.register_augment(client, augment.key, inst.activate)


def init(agents: list):
    for agent, _ in agents:
        agent.on_start()


async def step(client: object,
               agents: list,
               timeout: int,
               loop: BaseEventLoop):
    tasks = []

    for agent, agent_cfg in agents:
        tags = agent_cfg.tags

        if agent.lazy and not agent.ready:
            continue

        def event_fn(**kwargs):
            if "tags" in kwargs:
                for tag in tags:
                    kwargs["tags"].append(tag)
            else:
                kwargs["tags"] = tags

            if "time" not in kwargs:
                kwargs["time"] = int(timer())

            client.event(**kwargs)

        tasks.append(
            loop.create_task(
                agent.pull_metrics(event_fn, loop=loop)
            )
        )
    (done, pending) = await asyncio.wait(tasks, timeout=timeout, loop=loop)
    return (done, pending)


def instrumentation(client: processor.QClient,
                    interval: int,
                    delta: int,
                    events_count: int,
                    pending_events: int):
    send_heartbeat(
        event_fn=client.event,
        ttl=int(interval * 1.5)
    )
    send_timedelta(
        event_fn=client.event,
        td=delta,
        interval=interval
    )
    send_metrics_count(
        event_fn=client.event,
        count=events_count
    )
    send_pending_events_count(
        event_fn=client.event,
        count=events_count
    )


async def main_loop(cfg: Config,
                    transport_cls: Generic[T],
                    continue_fn: callable,
                    loop: BaseEventLoop):
    riemann = cfg.riemann
    transport = transport_cls(riemann.host, riemann.port)
    client = processor.QClient(transport)
    agents = create_agents(cfg.agents)
    register_augments(client, cfg.augments)
    executor = cfg.executor_class(max_workers=cfg.executors_count)
    loop.set_default_executor(executor)

    def handle_async_errors(loop, ctx):
        logger.error("Received error on async loop: {}"
                     .format(ctx["exception"]))

    init(agents)

    while True:
        ts = timer()
        (done, pending) = await step(client,
                                     agents,
                                     timeout=cfg.interval,
                                     loop=loop)
        logger.debug("Pending tasks: {}".format(pending))

        te = timer()
        td = int(te - ts)
        instrumentation(client=client,
                        interval=cfg.interval,
                        delta=td,
                        events_count=len(client.queue.events),
                        pending_events=len(pending))

        await processor.flush(client, transport)
        if continue_fn():
            await asyncio.sleep(cfg.interval - int(td), loop=loop)
        else:
            logger.info("Stopping Oshino")
            break

    client.on_stop()


def start_loop(cfg: Config, noop=False):
    logger.info("Initializing Oshino v{0}".format(get_version()))
    logger.info("Running forever in {0} seconds interval. Press Ctrl+C to exit"
                .format(cfg.interval))
    if cfg.sentry_dsn:
        try:
            client = SentryClient(cfg.sentry_dsn)
            logging.root.handlers.append(SentryHandler(client,
                                          level=logging.ERROR,
                                          bubble=True))
        except InvalidDsn:
            logger.warn("Invalid Sentry DSN '{0}' providen. Skipping"
                        .format(cfg.sentry_dsn))


    loop = create_loop()
    try:
        loop.run_until_complete(main_loop(cfg,
                                          cfg.riemann.get_transport(noop),
                                          forever,
                                          loop=loop))
    finally:
        loop.close()



import asyncio

from . import Agent


class StubAgent(Agent):

    def is_valid(self):
        return True

    async def process(self, event_fn):
        event_fn(metric_f=1.0,
                 service="test",
                 tags=["test"])


class LaggingAgent(Agent):

    def __init__(self, cfg, lag):
        super(LaggingAgent, self).__init__(cfg)
        self.lag = lag

    def is_valid(self):
        return True

    async def process(self, event_fn):
        await asyncio.sleep(self.lag)
        event_fn(metric_f=1.0,
                 service="lagging_test",
                 tags=["test"])
